"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const html_minifier_1 = require("html-minifier");
const jsdom_1 = require("jsdom");
const luxon_1 = require("luxon");
const tldjs_1 = require("tldjs");
const Resource_1 = require("../data/Resource");
const plugins_1 = require("../plugins");
const css_1 = require("../utils/css");
const element_1 = require("../utils/element");
const file_1 = require("../utils/file");
const twitter_1 = require("../utils/twitter");
class VJsdom {
    constructor(html, options = {}) {
        this.doctype = '<!DOCTYPE html>';
        const virtualConsole = new jsdom_1.VirtualConsole();
        this.dom = new jsdom_1.JSDOM(html, Object.assign({}, options, { virtualConsole }));
        const meta = this.metaElement;
        if (meta) {
            this.metadata = VJsdom.parseMetadata(meta);
        }
        virtualConsole.sendTo(console, { omitJSDOMErrors: true });
    }
    static parseMetadata(meta) {
        const parsed = JSON.parse(meta.content);
        return Object.assign({}, parsed, { _createdAt: luxon_1.DateTime.fromISO(parsed._createdAt) });
    }
    get document() {
        return this.dom.window.document;
    }
    get location() {
        return this.dom.window.location;
    }
    get metaElement() {
        return this.document.querySelector('meta[name=vanilla-clipper]');
    }
    insertMetadataToDocument() {
        if (!this.metadata) {
            this.metadata = {
                _version: 1,
                _createdAt: luxon_1.DateTime.local(),
                domain: tldjs_1.getDomain(this.location.href) || this.location.hostname,
                hostname: this.location.hostname,
                url: this.location.href,
                title: this.document.title,
            };
        }
        const meta = this.document.createElement('meta');
        meta.name = 'vanilla-clipper';
        meta.content = JSON.stringify(this.metadata);
        this.document.head.prepend(meta);
    }
    generate(withMetadata = false) {
        if (withMetadata) {
            this.insertMetadataToDocument();
        }
        const html = `${this.doctype}\n${this.document.documentElement.outerHTML}`;
        return {
            html: html_minifier_1.minify(html, { minifyJS: true, minifyCSS: true }),
            metadata: this.metadata,
        };
    }
    finder(...selectors) {
        return [...this.document.querySelectorAll(element_1.selectorsToString(...selectors))];
    }
    execPlugins() {
        plugins_1.jsdomPlugins.exec(this);
    }
    setElementAsRoot(selector) {
        const el = this.document.body.querySelector(selector);
        if (el) {
            this.document.body.innerHTML = el.outerHTML;
        }
    }
    getIframes() {
        return this.finder(element_1.buildVAttrSelector.iframeUuid());
    }
    replaceVideoElement(oldElement, url, isGif) {
        const el = this.document.createElement('video');
        el.dataset.vanillaClipperVideo = isGif ? 'gif' : 'video';
        el.src = url;
        el.style.width = '100%';
        el.style.height = '100%';
        el.style.position = 'relative';
        if (isGif) {
            el.setAttribute('muted', '');
        }
        el.autoplay = isGif;
        el.loop = isGif;
        el.controls = true;
        // oldElement.closest('.PlayableMedia-reactWrapper')!.replaceWith(el)
        oldElement.parentElement.parentElement.parentElement.replaceWith(el);
    }
    async embedTwitterVideo(tweetID) {
        const gifElement = this.document.querySelector('img[src*="/tweet_video_thumb/"]');
        const videoElement = this.document.querySelector('img[src*="/ext_tw_video_thumb/"]');
        if (gifElement) {
            const media = await twitter_1.getTwitterVideoURL(tweetID);
            if (!media)
                return;
            this.replaceVideoElement(gifElement, media.url, true);
        }
        if (videoElement) {
            const media = await twitter_1.getTwitterVideoURL(tweetID);
            if (!media)
                return;
            this.replaceVideoElement(videoElement, media.url, false);
        }
    }
    embedIFrameContents(iframeDataList) {
        return iframeDataList.map(({ uuid, html }) => {
            const iframe = this.finder(element_1.buildVAttrSelector.iframeUuid(uuid))[0];
            delete iframe.dataset.vanillaClipperIframeUuid;
            iframe.srcdoc = html;
            iframe.dataset.vanillaClipperSrc = iframe.src;
            iframe.removeAttribute('src');
            return iframe;
        });
    }
    async processResourcesInAttrs(noStoring = false) {
        await Promise.all([
            ...this.finder({ selector: '[src]', not: ['[src=""]', 'iframe'] }).map(async (el) => {
                el.removeAttribute('srcset');
                const url = el.dataset.vanillaClipperSrc || el.getAttribute('src');
                const dataURL = (el.dataset.vanillaClipperSrc && el.getAttribute('src')) || undefined;
                if (file_1.dataURLPattern.test(url)) {
                    return;
                }
                el.dataset.vanillaClipperSrc = url;
                if (noStoring) {
                    return;
                }
                const version = await Resource_1.Resource.store(this.location.href, url, dataURL);
                if (!version) {
                    return;
                }
                el.setAttribute('src', version.url);
            }),
            ...this.finder({
                selector: '[href]',
                not: [
                    '[href=""]',
                    'a',
                    'div',
                    '[rel~=alternate]',
                    '[rel~=canonical]',
                    '[rel~=prev]',
                    '[rel~=next]',
                ],
            }).map(async (el) => {
                const url = el.dataset.vanillaClipperHref || el.getAttribute('href');
                const dataURL = (el.dataset.vanillaClipperHref && el.getAttribute('href')) || undefined;
                if (file_1.dataURLPattern.test(url)) {
                    return;
                }
                el.dataset.vanillaClipperHref = url;
                if (noStoring) {
                    return;
                }
                const version = await Resource_1.Resource.store(this.location.href, url, dataURL);
                if (!version) {
                    return;
                }
                el.setAttribute('href', version.url);
            }),
        ]);
    }
    async optimizeStylesInShadowDOM() {
        await Promise.all(this.finder(element_1.buildVAttrSelector.shadowContent()).map(async (host) => {
            if (!host.dataset.vanillaClipperShadowContent) {
                return;
            }
            const shadow = host.attachShadow({ mode: 'open' });
            shadow.innerHTML = host.dataset.vanillaClipperShadowContent;
            const cssTexts = [];
            shadow.querySelectorAll('style').forEach(el => {
                cssTexts.push(el.innerHTML);
                el.remove();
            });
            const optimized = await Promise.all(cssTexts.map(text => css_1.optimizeCSS({ text, url: this.location.href }), shadow));
            this.appendStyleSheets(optimized, shadow);
            host.dataset.vanillaClipperShadowContent = shadow.innerHTML;
        }));
    }
    appendStyleSheets(cssTexts, shadowRoot) {
        const styleElements = cssTexts.map(text => {
            const el = this.document.createElement('style');
            el.dataset.vanillaClipperStyle = '';
            el.innerHTML = text;
            return el;
        });
        (shadowRoot ? shadowRoot : this.document.head).append(...styleElements);
    }
    appendScriptToHead() {
        const shadowHostSelector = element_1.buildVAttrSelector.shadowContent();
        const videoElementSelector = element_1.buildVAttrSelector.video();
        async function main() {
            if (document.readyState === 'loading') {
                document.addEventListener('DOMContentLoaded', onload);
            }
            else {
                onload();
            }
            function onload() {
                const videoElement = document.querySelector(videoElementSelector);
                if (videoElement) {
                    const playOrPause = () => videoElement.paused ? videoElement.play() : videoElement.pause();
                    videoElement.onclick = playOrPause;
                    document.body.onkeypress = () => {
                        if (window.event.keyCode === 32) {
                            ;
                            window.event.preventDefault();
                            playOrPause();
                        }
                    };
                }
                const shadowHosts = document.querySelectorAll(shadowHostSelector);
                shadowHosts.forEach(host => {
                    if (!host.dataset.vanillaClipperShadowContent) {
                        return;
                    }
                    const shadow = host.attachShadow({ mode: 'open' });
                    shadow.innerHTML = host.dataset.vanillaClipperShadowContent;
                });
            }
        }
        const script = this.document.createElement('script');
        script.dataset.vanillaClipperScript = '';
        script.innerHTML = `
            const shadowHostSelector = '${shadowHostSelector}'
            const videoElementSelector = '${videoElementSelector}'

            ${main.toString()}
            main()
        `;
        this.document.head.appendChild(script);
    }
}
exports.VJsdom = VJsdom;
//# sourceMappingURL=VJsdom.js.map