"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_extra_1 = require("fs-extra");
const luxon_1 = require("luxon");
require("mime-types");
const mime_types_1 = require("mime-types");
const ulid_1 = require("ulid");
const url_1 = require("url");
const config_store_1 = require("../config-store");
const VDB_1 = require("../core/VDB");
const utils_1 = require("../utils");
const file_1 = require("../utils/file");
function resourcePath(...pathSegments) {
    return file_1.dataPath('resources', ...pathSegments);
}
const db = new VDB_1.VDB(file_1.dataPath('resources.json'), true, false);
class Resource {
    constructor(url) {
        this.url = url;
        this.versions = [];
    }
    recentVersion() {
        const last = this.versions[this.versions.length - 1];
        if (last && -1 < luxon_1.DateTime.fromISO(last.createdAt).diffNow('hours').hours) {
            return last;
        }
    }
    save() {
        db.put(this.url, { url: this.url, versions: this.versions });
    }
    async createVersion({ buffer, hash, mimetype, }) {
        const outputDir = luxon_1.DateTime.local().toFormat('yyyyMMdd');
        const id = ulid_1.ulid();
        const ext = mimetype && mime_types_1.extension(mimetype);
        const filename = ext ? `${id}.${ext}` : id;
        const path = `${outputDir}/${filename}`;
        await fs_extra_1.outputFile(resourcePath(outputDir, filename), buffer);
        if (hash) {
            const version = {
                createdAt: utils_1.now(),
                hash,
                path,
            };
            this.versions.push(version);
            this.save();
            return version;
        }
        return { path };
    }
    static get(url) {
        const resource = new Resource(url);
        try {
            const data = db.get(url);
            Object.assign(resource, data);
        }
        catch (error) { }
        return resource;
    }
    static withRelativeURL(version) {
        if (!version) {
            return;
        }
        return { version, url: `../../resources/${version.path}` };
    }
    static async store(baseURL, relativeURL, dataURL) {
        const bufferFromDataURL = file_1.dataURLToBuffer(relativeURL);
        if (bufferFromDataURL) {
            const resource = new Resource(relativeURL);
            return this.withRelativeURL(await resource.createVersion(bufferFromDataURL));
        }
        const url = url_1.resolve(baseURL, relativeURL);
        const resource = this.get(url);
        const recentVersion = resource.recentVersion();
        if (recentVersion) {
            return this.withRelativeURL(recentVersion);
        }
        try {
            const response = await (async () => {
                if (dataURL) {
                    return file_1.dataURLToBuffer(dataURL);
                }
                const { headers: { 'content-type': mimetype, 'content-length': size }, } = await utils_1.got.head(dataURL || url);
                if (size && parseInt(size) > config_store_1.config.resource.maxSize) {
                    return;
                }
                const { body } = await utils_1.got.get(url, { encoding: null });
                if (body.length > config_store_1.config.resource.maxSize) {
                    return;
                }
                return { buffer: body, mimetype };
            })();
            if (!response) {
                return;
            }
            const hash = file_1.getHash(response.buffer);
            const matchedVersion = resource.versions.find(v => v.hash === hash);
            if (matchedVersion) {
                return this.withRelativeURL(matchedVersion);
            }
            return this.withRelativeURL(await resource.createVersion({
                buffer: response.buffer,
                hash,
                mimetype: response.mimetype,
            }));
        }
        catch (error) {
            utils_1.sig.error(error);
            return;
        }
    }
}
exports.Resource = Resource;
//# sourceMappingURL=Resource.js.map