"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const rain_1 = require("@yarnaimo/rain");
const css_tree_1 = __importDefault(require("css-tree"));
const url_1 = require("url");
const _1 = require(".");
const Resource_1 = require("../data/Resource");
const file_1 = require("./file");
const doubleQuote = '"';
const singleQuote = "'";
exports.cssURLPattern = /([:,\s]\s*url)\s*\((?!\s*data:.+,)\s*(\S+?)\s*\)/gi;
const fontFormats = {
    woff2: 'woff2',
    woff: 'woff',
    ttf: 'truetype',
    otf: 'opentype',
};
const fontFormatNames = Object.values(fontFormats);
const unescapeURL = (url) => url.startsWith(doubleQuote) && url.endsWith(doubleQuote)
    ? url.slice(1, -1).replace(/\\"/g, doubleQuote)
    : url.startsWith(singleQuote) && url.endsWith(singleQuote)
        ? url.slice(1, -1).replace(/\\'/g, singleQuote)
        : url;
async function extractOrFetchCSS(sheetDataList) {
    const sheets = await rain_1.Rarray.waitAll(sheetDataList, async (data) => {
        try {
            if (data.type === 'error') {
                throw data.error;
            }
            if (data.type === 'text') {
                return { text: data.text, url: data.url };
            }
            const { body } = await _1.got.get(data.url);
            return { text: body, url: data.url };
        }
        catch (_error) {
            return { text: _1.commentOutError(_error), url: data.url };
        }
    });
    return sheets;
}
exports.extractOrFetchCSS = extractOrFetchCSS;
function extractFontsFromValue(value) {
    const local = [];
    const remote = [];
    const chunks = value.children.toArray().reduce((array, node) => {
        if (node.type === 'Operator' && node.value === ',') {
            return [...array, []];
        }
        else {
            array[array.length - 1].push(node);
            return array;
        }
    }, [[]]);
    chunks.forEach(chunk => {
        const functionNodes = chunk.filter((node) => node.type === 'Function');
        const localNode = functionNodes.find(node => node.name === 'local');
        const formatNode = functionNodes.find(node => node.name === 'format');
        const urlNode = chunk.find((node) => node.type === 'Url' && !file_1.dataURLPattern.test(node.value.value));
        if (localNode) {
            local.push(localNode);
            return;
        }
        let formatName;
        if (formatNode) {
            const data = formatNode.children.first();
            if (data && data.type === 'String') {
                const _formatName = data.value.replace(/['"]/g, '');
                if (fontFormatNames.includes(_formatName)) {
                    formatName = _formatName;
                }
            }
        }
        else if (urlNode) {
            const ext = file_1.extractExtensionFromURL(urlNode.value.value);
            if (ext && ext in fontFormats) {
                formatName = fontFormats[ext];
            }
        }
        if (urlNode && formatName) {
            remote.push({ path: urlNode.value.value, formatName });
        }
    });
    return { local, remote };
}
function processSelectorList(prelude) {
    prelude.children.forEach((node, item, list) => {
        if (node.type === 'Selector') {
            node.children.forEach((node, item, list) => {
                if (node.type === 'Percentage') {
                    throw new Error();
                }
                if (node.type === 'PseudoElementSelector') {
                    list.remove(item);
                }
                if (node.type === 'PseudoClassSelector' &&
                    [
                        'active',
                        'checked',
                        'disabled',
                        'focus',
                        'hover',
                        'link',
                        'required',
                        'visited',
                        'after',
                        'before',
                        'cue',
                        'first-letter',
                        'first-line',
                        'selection',
                        'slotted',
                    ].includes(node.name)) {
                    list.remove(item);
                }
            });
            if (!node.children.getSize()) {
                list.remove(item);
            }
        }
    });
}
async function replaceImports(ast, baseURL) {
    let promises = [];
    css_tree_1.default.walk(ast, (node, item, list) => {
        if (node.type === 'Atrule' &&
            node.name === 'import' &&
            node.prelude &&
            node.prelude.type === 'AtrulePrelude') {
            const urlNode = node.prelude.children
                .toArray()
                .find((n) => n.type === 'Url');
            if (!urlNode) {
                return;
            }
            const url = url_1.resolve(baseURL, unescapeURL(urlNode.value.value));
            promises.push((async () => {
                const { body } = await _1.got.get(url).catch(() => ({ body: '' }));
                const parsed = css_tree_1.default.parse(body);
                if (parsed.type !== 'StyleSheet') {
                    return;
                }
                await replaceImports(parsed, url);
                list.replace(item, parsed.children);
            })());
        }
    });
    await Promise.all(promises);
}
async function optimizeCSS({ text, url: baseURL }, document, noStoring = false) {
    const ast = css_tree_1.default.parse(text);
    await replaceImports(ast, baseURL);
    css_tree_1.default.walk(ast, (node, item, list) => {
        if (document && node.type === 'Rule' && node.prelude.type === 'SelectorList') {
            const includesURL = node.block.children.some(node => {
                if (node.type !== 'Declaration' || node.value.type === 'Raw') {
                    return false;
                }
                return node.value.children.some(node => node.type === 'Url');
            });
            if (!includesURL) {
                return;
            }
            try {
                const prelude = css_tree_1.default.clone(node.prelude);
                processSelectorList(prelude);
                if (!prelude.children.getSize()) {
                    return;
                }
                const selector = css_tree_1.default.generate(prelude);
                const element = document.querySelector(selector);
                if (!element) {
                    list.remove(item);
                }
            }
            catch (error) { }
        }
    });
    let promises = [];
    css_tree_1.default.walk(ast, node => {
        if (node.type === 'Atrule' && node.name === 'font-face' && node.block) {
            const base = [];
            const local = [];
            const remote = [];
            node.block.children.toArray().forEach(line => {
                if (line.type === 'Declaration' &&
                    line.property === 'src' &&
                    line.value.type === 'Value') {
                    const fonts = extractFontsFromValue(line.value);
                    local.push(...fonts.local);
                    remote.push(...fonts.remote);
                }
                else {
                    base.push(line);
                }
            });
            const optimalRemote = remote.sort((a, b) => {
                return ((fontFormatNames.indexOf(a.formatName) + 1 || 9) -
                    (fontFormatNames.indexOf(b.formatName) + 1 || 9));
            })[0];
            const optimalRemoteNodes = optimalRemote
                ? [
                    {
                        type: 'Url',
                        value: { type: 'Raw', value: optimalRemote.path },
                    },
                    { type: 'WhiteSpace', value: ' ' },
                    {
                        type: 'Function',
                        name: 'format',
                        children: new css_tree_1.default.List().fromArray([
                            { type: 'String', value: `'${optimalRemote.formatName}'` },
                        ]),
                    },
                ]
                : [];
            const chunks = [...local.map(l => [l]), optimalRemoteNodes];
            const children = chunks.reduce((list, chunk) => {
                if (list.getSize()) {
                    list.appendData({ type: 'Operator', value: ',' });
                }
                chunk.forEach(node => list.appendData(node));
                return list;
            }, new css_tree_1.default.List());
            node.block.children.fromArray([
                {
                    type: 'Declaration',
                    property: 'src',
                    important: false,
                    value: { type: 'Value', children },
                },
                ...base,
            ]);
        }
        if (node.type === 'Url') {
            const { value: url } = node.value;
            const relativeURL = unescapeURL(url);
            if (file_1.dataURLPattern.test(relativeURL) || noStoring) {
                return;
            }
            promises.push((async () => {
                const version = await Resource_1.Resource.store(baseURL, relativeURL);
                if (!version) {
                    return;
                }
                node.value.value = version.url;
            })());
        }
    });
    await Promise.all(promises);
    const generated = css_tree_1.default.generate(ast);
    return generated;
}
exports.optimizeCSS = optimizeCSS;
//# sourceMappingURL=css.js.map